import torch
#import numpy as np

#from typing import Union
from scipy.spatial import KDTree
from scipy.sparse.csgraph import dijkstra
#from scipy.stats import shapiro, combine_pvalues    
#from pingouin import multivariate_normality
#from tqdm import tqdm
#from LAMINAR.Flow.planarCNF import PlanarCNF, train_PlanarCNF
from LAMINAR.Flow.OTFlow import Phi, train_OTFlow, integrate
from LAMINAR.utils.gaussian2uniform import gaussian_to_sphere
from LAMINAR.utils.geodesics import geodesic_length, geodesic_path, geodesic_straight_line

'''
Implementation of the LAM algorithm using a normalizing flow to transform the data
'''
class LAMINAR():
    def __init__(self,
                 data, 
                 alph = [1.0, 100.0, 5.0],
                 nt = 8, 
                 nt_val = 8,
                 nTh = 3,
                 m = 32,
                 lr = 0.1,
                 drop_freq = 100,
                 lr_drop = 2,
                 k_neigh = 10,
                 epochs = 1500):
        
        self.device = data.device
        self.data = data

        self.alph = alph
        self.nt = nt
        self.nt_val = nt_val
        self.nTh = nTh
        self.m = m
        self.lr = lr
        self.drop_freq = drop_freq
        self.lr_drop = lr_drop
        self.k_neigh = k_neigh 
        self.epochs = epochs

        self.d = self.data.shape[1]
        self.n = self.data.shape[0]

        # split the data into training and validation
        self.data_train = self.data[:int(self.n*0.8)]
        self.data_val = self.data[int(self.n*0.8):]

        # initialize the normalizing flow
        self.net = Phi(self.nTh, self.m, self.d, alph=self.alph, device=self.device)
        self.optimizer = torch.optim.Adam(self.net.parameters(), lr=self.lr)

        # train the model
        self.loss_hist = train_OTFlow(self.net, self.optimizer, self.data_train, self.data_val, self.epochs, self.nt, self.nt_val, self.drop_freq, self.lr_drop)

        # set up the graph
        self.set_up_graph()


    def switch_device(self, device):
        self.device = device
        self.net.to(device)
    

    def set_up_graph(self):
        self.X_pushed = integrate(self.data, self.net, [0, 1], nt=self.nt, stepper="rk4", alph=self.alph, intermediates=False).cpu().detach()[:, :self.d]
        self.X_pushed = gaussian_to_sphere(self.X_pushed)

        kdt = KDTree(self.X_pushed)
        dists, neighs = kdt.query(self.X_pushed, k=self.k_neigh)

        # get all start points
        starts = self.data[neighs[:, 0]]
        # repeat every point in starts for k_neigh times so that the same point appears k_neigh times right after each other
        starts = starts.repeat_interleave(self.k_neigh, dim=0)

        ends = self.data[neighs.flatten()]

        distances = geodesic_straight_line(starts, ends, self.net.metric_tensor, inbetween=1)

        row_indices = torch.arange(neighs.shape[0]).repeat_interleave(neighs.shape[1])
        col_indices = neighs.flatten()
        dist_values = distances.flatten()

        self.graph = torch.zeros(self.n, self.n).to(self.device)
        self.graph[row_indices, col_indices] = dist_values

        # graph symmetric, by transposition and insertion of values which are not yet in the graph
        graph_sub_transpose = self.graph - self.graph.t()

        # set positive values to zero
        graph_sub_transpose[graph_sub_transpose > 0] = 0
        self.graph = self.graph - graph_sub_transpose

        self.dist_matrix, self.predecessors = dijkstra(self.graph.detach().cpu().numpy(), return_predecessors=True)
        

    def expand_graph(self, additional_points):
        # additional_points is an array of shape (m, d) of points which temporarily need to be added to the graph
        # returns the expanded graph, the distance matrix and the predecessors

        #additional_points = additional_points.reshape(-1, self.d)
        expanded_data = torch.cat([self.data, additional_points], dim=0)
        additional_points_pushed = integrate(additional_points, self.net, [0, 1], nt=self.nt_val, stepper="rk4", alph=self.alph, intermediates=False).cpu().detach()[:, :self.d]
        additional_points_pushed = gaussian_to_sphere(additional_points_pushed)

        expanded_data_pushed = torch.concatenate([self.X_pushed, additional_points_pushed], dim=0)

        kdt = KDTree(expanded_data_pushed)
        _, neighs = kdt.query(additional_points_pushed, k=self.k_neigh)

        starts = additional_points.repeat_interleave(self.k_neigh, dim=0)
        ends = expanded_data[neighs.flatten()]

        distances = geodesic_straight_line(starts, ends, self.net.metric_tensor, inbetween=1)

        row_indices = torch.arange(neighs.shape[0]).repeat_interleave(neighs.shape[1])
        col_indices = neighs.flatten()

        expanded_graph = torch.zeros(expanded_data.shape[0], expanded_data.shape[0]).to(self.device)
        expanded_graph[:self.n, :self.n] = self.graph
        expanded_graph[row_indices + self.n, col_indices] = distances

        # graph
        expanded_graph_sub_transpose = expanded_graph - expanded_graph.t()
        expanded_graph_sub_transpose[expanded_graph_sub_transpose > 0] = 0
        expanded_graph = expanded_graph - expanded_graph_sub_transpose

        dist_matrix, predecessors = dijkstra(expanded_graph.detach().cpu().numpy(), return_predecessors=True)

        return expanded_graph, dist_matrix, predecessors 


    def check_expansion(self, points):
        # for each point, check wether it is in the data, if so note the indices
        # else add the point to the graph

        # check if the points are in the data
        kdt = KDTree(self.data)
        dists, idx = kdt.query(points, k=1)

        dists = torch.tensor(dists)

        # args where dists is zero
        in_data = torch.where(dists == 0)[0]
        at_indices = idx[in_data]
        at_indices = torch.tensor(at_indices).reshape(-1)

        not_in_data = torch.where(dists != 0)[0]

        # expand graph by not_in_data
        # if not_in_data is not empty
        if not_in_data.shape[0] != 0:
            # expand the graph
            _, dist_matrix, predecessors = self.expand_graph(points[not_in_data])

            # note the indices
            not_in_data_idx = torch.arange(self.n, self.n + not_in_data.shape[0]) #[not_in_data]

            # concat the windices of points in the data
            idx = torch.cat([at_indices, not_in_data_idx])

            # make the distance matrix a tensor and return
            dist_matrix = torch.tensor(dist_matrix)

            return idx, dist_matrix, predecessors
        
        else:
            # no extension needed, just return the indices and the distance matrix
            idx = at_indices
            dist_matrix = torch.tensor(self.dist_matrix)

            return idx, dist_matrix, self.predecessors

    
    def query(self, start, k=None):     # TODO expand to add point besides the data
        # start is an array of shape (m, d)
        # if shape is (d,) reshape to (1, d)
        if start.shape == (self.d,):
            start = start.reshape(1, self.d)

        # calculate the k nearest points and their distance for each start point
        # expand the graph with the new points
        idx_points, dist_matrix, _ = self.check_expansion(start)
        dists = torch.tensor(dist_matrix[idx_points]) # shape (m, n)
        dists, idx = torch.sort(dists, dim=1)
    
        if k is not None:
            return idx[:, :k], dists[:, :k].detach()
        
        else:
            return idx, dists.detach()
    
    def distance_approx(self, start, end, return_path=False):
        # expand graph by end and start points
        start = start.unsqueeze(0) if start.dim() == 1 else start
        end = end.unsqueeze(0) if end.dim() == 1 else end
        all_points = torch.cat([self.data, start, end], dim=0)
        #_, dist_matrix, predecessors = self.expand_graph(torch.cat([start, end], dim=0))

        idx, dist_matrix, predecessors = self.check_expansion(torch.cat([start, end], dim=0))

        start_idx = idx[-2]
        end_idx = idx[-1]    

        if return_path:
            path_idx = [end_idx]
            current = end_idx

            while current != start_idx:
                current = predecessors[start_idx, current]
                path_idx.append(current)

            path_idx = torch.tensor(path_idx).flip(0)
            path = all_points[path_idx]

            #dist = geodesic_length(path.reshape(1, path.shape[0], self.d), start, end, self.net.metric_tensor)
            dist = dist_matrix[start_idx, end_idx]

            return dist, path
        
        else:
            dist = dist_matrix[start_idx, end_idx]

            return dist


    def distance_smooth(self, start, end):
        _, path = self.distance_approx(start, end)
        points, _ = geodesic_path(start, end, self.net.metric_tensor, lr=1e-2, initial_guess=path, max_iter=1000)
        dist = geodesic_length(points[1:-1].reshape(1, points[1:-1].shape[0], self.d), start, end, self.net.metric_tensor)

        return dist[0].detach(), points.detach()
    


import torch
from torch.func import vmap


def christoffel_symbol(x, metric_func, eps=1e-6, numeric_diff = True):
    '''
    Calculate the christoffel symbols at the locations x (shape: n, dim) for the metric given by metric_func
    '''

    device = x.device
    dim = x.shape[1] # dimension of the space
    n = x.shape[0] # number of points

    # make input (shape: n, dim+1, dim) for metric call, which calls for every x and a small deviation in every direction for every x
    if numeric_diff:
        x_in = x.unsqueeze(1).repeat(1, 2*dim+1, 1)
        # add and subtract small deviation in every direction
        x_in[:, 1:dim+1] += torch.eye(dim, device=device).unsqueeze(0) * eps
        x_in[:, dim+1:] -= torch.eye(dim, device=device).unsqueeze(0) * eps

        # calculate metric
        #print(x_in.shape)
        g = metric_func(x_in)

        # get gradients of g wrt x (shape: n, dim, dim, dim)
        g_grad1 = (g[:, 1:dim+1] - g[:, 0].repeat(1, dim, 1, 1).reshape(n, dim, dim, dim)) / eps
        g_grad2 = (g[:, dim+1:] - g[:, 0].repeat(1, dim, 1, 1).reshape(n, dim, dim, dim)) / -eps

        g_grad = (g_grad1 + g_grad2) / 2

        g = g[:, 0] # only keep the metric at the point

    else:
        x_in = x
        g = metric_func(x_in)

        jacobian_func = torch.func.jacrev(metric_func, argnums=0)
        batched_jacobian_func = vmap(jacobian_func, in_dims=0)
        g_grad = batched_jacobian_func(x).permute(0, 3, 1, 2)

        #print(g_grad.shape)
        
    inv_g = torch.inverse(g)

    # calculate christoffel symbols

    term1 = g_grad.permute(0, 3, 1, 2)
    term2 = g_grad.permute(0, 3, 2, 1)
    term3 = g_grad.permute(0, 1, 2, 3)

    sum_term = (term1 + term2 - term3) / 2

    christoffel = torch.einsum('nij,njkl->nikl', inv_g, sum_term)

    return christoffel

def geodesic_equation(path, metric_func, eps=1e-6):

    # velocity at each point
    v = (path[2:] - path[:-2])/2 # shape: n-2, dim

    # acceleration at each point
    a = path[2:] - 2*path[1:-1] + path[:-2] # shape: n-2, dim

    # calculate christoffel symbols at each point
    christoffel = christoffel_symbol(path[1:-1], metric_func, eps=eps, numeric_diff=True) # shape: n-2, dim, dim, dim

    # calculate the geodesic equation
    delta_mu = torch.einsum('ndij,ni,nj->nd', christoffel, v, v) # shape: n-2, dim
    delta_mu = a + delta_mu

    tot_deviation = torch.sum(torch.sqrt(torch.sum(delta_mu**2, dim=1)))

    return tot_deviation


def geodesic_length(points, start, end, metric_func):
    # points is a tesor of nex-by-steps-by-d
    # allow calculations of multiple paths with points as intermediate points, while start and end are fixed
    n_paths = points.shape[0]

    # concat start, points, end
    paths = torch.cat([start.repeat(n_paths, 1, 1), points, end.repeat(n_paths, 1, 1)], dim=1)

    # calculate delta of points
    delta_x = paths[:, 1:] - paths[:, :-1]

    g = metric_func((paths[:, 1:]+paths[:, :-1])/2)

    ds_squared = torch.einsum('abi,abij,abj->ab', delta_x, g, delta_x)

    total_length = torch.sqrt(ds_squared).sum(dim=1)

    variance = torch.var(torch.sqrt(ds_squared))
    return total_length, variance


def geodesic_straight_line(starts, ends, metric_func, inbetween = 10):
    # starts shape (n, d)
    # ends shape (n, d)

    n = starts.shape[0]
    d = starts.shape[1]

    # for every instance n, make a straight line between start and end
    points = torch.linspace(0, 1, inbetween+2).view(-1, 1).repeat(n, 1, 1)
    points = torch.einsum('ij,ikj->ikj', (ends - starts), points.repeat(1, 1, d).reshape(n, inbetween+2, 2)) + starts.unsqueeze(1)

    delta_x = points[:, 1:] - points[:, :-1]
    g = metric_func((points[:, 1:] + points[:, :-1]) / 2)

    ds_squared = torch.einsum('abi,abij,abj->ab', delta_x, g, delta_x)
    total_length = torch.sqrt(ds_squared).sum(dim=1)

    return total_length


def geodesic_path(start, end, metric_func, inbetween = 8, lr = 1e-2, initial_guess = None, max_iter = 1000):
    device = start.device

    if initial_guess is None:
        points = torch.linspace(0, 1, inbetween+2).to(device).view(-1, 1).to(device) * (end - start) + start
        points = points[1:-1]
    
    else:
        points = initial_guess[1:-1]
        points = points.to(device)

    points.requires_grad = True

    opt = torch.optim.Adam([points], lr=lr)

    loss_hist = []

    best_loss = 9999999
    lr_drop_count = 0
    best_points = points

    for _ in range(max_iter):
        opt.zero_grad()

        path = torch.concatenate([start.reshape(1, 2), points, end.reshape(1, 2)], dim=0)
        loss = geodesic_equation(path, metric_func)

        loss.backward()       

        opt.step()

        loss_hist.append(loss.item())

        if loss.item() < best_loss:
            best_loss = loss.item()
            best_points = points.detach()
            lr_drop_count = 0

        else:
            lr_drop_count += 1

        if lr_drop_count > 50:
            # drop learning rate and reset points to best points
            lr_drop_count = 0
            opt = torch.optim.Adam([points], lr=lr/2)
            points = best_points
            points.requires_grad = True
            
        if _ % 100 == 0:
            print(f"Iteration {_} | Loss: {loss.item()}")
        
    print('Max iterations reached')
    print(f"Final loss: {best_loss}")

    points = torch.concatenate([start.reshape(1, 2), best_points, end.reshape(1, 2)], dim=0)

    return points, loss_hist


def geodesic_path2(start, end, metric_func, inbetween = 8, lr = 1e-2, initial_guess = None, max_iter = 1000):
    device = start.device

    if initial_guess is None:
        points = torch.linspace(0, 1, inbetween+2).to(device).view(-1, 1).to(device) * (end - start) + start
        points = points[1:-1]
    
    else:
        points = initial_guess[1:-1]
        points = points.to(device)

    points.requires_grad = True

    opt = torch.optim.Adam([points], lr=lr)

    loss_hist = []

    best_loss = 9999999
    lr_drop_count = 0
    best_points = points

    for _ in range(max_iter):
        opt.zero_grad()

        path = torch.concatenate([start.reshape(1, 2), points, end.reshape(1, 2)], dim=0)
        loss = geodesic_equation(path, metric_func)

        loss.backward()       

        opt.step()

        loss_hist.append(loss.item())

        if loss.item() < best_loss:
            best_loss = loss.item()
            best_points = points.detach()
            lr_drop_count = 0

        else:
            lr_drop_count += 1

        if lr_drop_count > 50:
            # drop learning rate and reset points to best points
            lr_drop_count = 0
            opt = torch.optim.Adam([points], lr=lr/2)
            #points = best_points
            #points.requires_grad = True
            
        if _ % 100 == 0:
            print(f"Iteration {_} | Loss: {loss.item()}")
        
    print('Max iterations reached')
    print(f"Final loss: {best_loss}")

    points = torch.concatenate([start.reshape(1, 2), best_points, end.reshape(1, 2)], dim=0)

    return points, loss_hist